#Eneroth Upright Extruder

#Author:	Julia Christina Eneroth, Eneroth3, eneroth3@gmail.com

module Ene_uprightExtruder

#Module variables
@vectorUpright = Geom::Vector3d.new(0,0,1)

#Menu
menu = UI.menu("Plugins").add_submenu("保持直立跟随")
menu.add_item("保持直立跟随") { self.userInput }
menu.add_item("设置 \"Upright\" 轴向（根据坐标轴）") { self.setUpright }
menu.add_item("设置 \"Upright\" 轴向（根据选择线）") { self.setUptightFromEdge }

#Toolbar
tb = UI::Toolbar.new "保持直立跟随"

cmd = UI::Command.new("保持直立跟随") { self.userInput }
cmd.large_icon = "extrude.png"
cmd.small_icon = "extrude_small.png"
cmd.tooltip = "保持直立跟随"
cmd.status_bar_text = "将选择面沿选择线放样同时保持放样截面直立"
tb.add_item cmd

cmd = UI::Command.new("设置 \"Upright\" 轴向（根据坐标轴）") { self.setUpright }
cmd.large_icon = "vectorByCoords.png"
cmd.small_icon = "vectorByCoords_small.png"
cmd.tooltip = "设置 \"Upright\" 轴向（根据坐标轴）"
cmd.status_bar_text = "设置 \"upright\" 某坐标轴作为放样时保持直立的轴向。"
tb.add_item cmd

cmd = UI::Command.new("设置 \"Upright\" 轴向（根据选择线）") { self.setUptightFromEdge }
cmd.large_icon = "vectorFromEdge.png"
cmd.small_icon = "vectorFromEdge_small.png"
cmd.tooltip = "设置 \"Upright\" 轴向（根据选择线）"
cmd.status_bar_text = "设置\"upright\" 选择线作为放样时保持直立的轴向。"
tb.add_item cmd

UI.start_timer(0.1, false){ tb.restore }#Use timer as workaround for bug 2902434.

def self.flattenVector(vector,plane)
	#Flatten vector to plane
	#plane's point does not affect output, only its normal does
	
	pointStart = plane[0]
	normal = plane[1]
	pointVectorEnd = pointStart.offset vector#Point at end of vector
	lineToPlane = [pointVectorEnd, normal]#Line from previous point towards plane
	pointIntesection = Geom.intersect_line_plane lineToPlane, plane#Point where previous line intersects vector
	outputVector = pointIntesection.- pointStart#vector from startpoint to intersection
	return outputVector
end#def

def self.removeOrSoftenSmooth(edge)
	#Removes edge if bounded faces are co-planar, otherwise soften
	if edge.faces.length == 2
		onPlane = true
		edge.faces[1].vertices.each do |i|
			#Use classify_point instead of normal vector to see if face is co-planar. More precise
			onPlane = false if (edge.faces[0].classify_point i.position) == Sketchup::Face::PointNotOnPlane
		end#each
		if onPlane
			edge.erase!
		else
			edge.soft = edge.smooth = true
		end#if
	end#if
end#def

def self.userInput
	#User interface
	#In its own definition in so extruder itself can be called from other scripts
	
	#Selection is used as input
	#Must contain one face that is to be extruded
	#Must contain connected lines used as path

	#Get input from selection
	edges = []#Edges forming path
	faces = []#Face to extrude
	Sketchup.active_model.selection.each do |i|
		edges << i if i.class == Sketchup::Edge
		faces << i if i.class == Sketchup::Face
	end#each
	
	#Validate input
	if faces.length == 0
		UI.messagebox("请选择一个面以放样")
		return
	end#if
	if faces.length > 1
		UI.messagebox("无法同时对多个面进行放样")
		return
	end#if
	edges.delete_if{|i| i.faces.include? faces[0]}
	if edges.length == 0
		UI.messagebox("请选择一个或多个连接的边线，以进行放样")
		return
	end#if
	
	#Path
	path = []
	#Indexes of edges already put to path array
	edgesUsed = []
	#Add coordinates of first edge to path
	path << edges[0].start.position
	path << edges[0].end.position
	edgesUsed << 0
	#Double loop edges
	(0..(edges.length-1)).each do
		(0..(edges.length-1)).each do |i|
			#If edge is not already used and connects to an end of path, add edge's other point to that end
			unless edgesUsed.include? i
				if edges[i].start.position == path[0]
					path.unshift edges[i].end.position
					edgesUsed << i
				elsif edges[i].end.position == path[0]
					path.unshift edges[i].start.position
					edgesUsed << i
				elsif edges[i].start.position == path[-1]
					path << edges[i].end.position
					edgesUsed << i
				elsif edges[i].end.position == path[-1]
					path << edges[i].start.position
					edgesUsed << i
				end#if elsif
			end#unless
		end#each
	end#each
	
	#Validate path
	unless edgesUsed.length == edges.length
		UI.messagebox("请选择有效的线进行放样")
		return
	end#unless
	
	#Start operation
	Sketchup.active_model.start_operation "Extrude Path"
	
	#Call extrude function
	self.extrude(faces[0], path, @vectorUpright)
	Sketchup.status_text= "Face has been extruded."
	
	#End operation
	Sketchup.active_model.commit_operation
	
end#def

def self.setUpright
	#Set "upright" vector by coordinates
	
	prompts = ["X (红)", "Y (绿)", "Z (蓝)"]
	defaults = @vectorUpright.to_a
	title = "设置 \"Upright\" 轴向"
	input = UI.inputbox prompts, defaults, title
	return unless input
	vector = Geom::Vector3d.new
	(0..2).each do |i|
		vector[i] = input[i]
	end#each
	unless vector.valid?
		UI.messagebox("轴向长度不能为0.")
		return
	end#unless
	@vectorUpright = vector
	Sketchup.status_text= "\"Upright\" 轴向已经被设置."
end#def

def self.setUptightFromEdge
	#Set "upright" vector from edge
	
	ss = Sketchup.active_model.selection
	unless ss.length == 1 and ss[0].class == Sketchup::Edge
		UI.messagebox("必须至少选择一个直线")
		return
	end#unless
	@vectorUpright = ss[0].line[1]
	Sketchup.status_text= "\"Upright\" 轴向已经被设置."
end#def

def self.extrude(extrusion, path, vectorUpright=nil)
	#Actual extrude function.
	#Either called from userInput or external script.
	#Extrusion is Face face about to be extruded, path an Array of Points3ds to follow, vectorUpright is the vector to align extrusion to (by default upright)
	#If path start and ends with same point it's considered closed.
	#If not closed, the face is extruded relative to the nearest end of the path.
	#If closed, the face is extruded relative to its nearest point in the path.
	
	#Set vectorUpright to vertical if not set
	vectorUpright = Geom::Vector3d.new(0,0,1) unless vectorUpright

	#Validate input
	if vectorUpright.class != Geom::Vector3d
		UI.messagebox("无有效的轴向设置。")
		return
	end#if
	if extrusion.class != Sketchup::Face
		UI.messagebox("无有效的放样面")
	end#if
	
	#Check whether path is closed
	closed = (path[0] == path[-1])
	
	#Get drawing context of extrusion face(group, componentdefinition or model)
	drwingContext = extrusion.parent
	ents = drwingContext.entities
	
	#Get material of extrusion face
	mat = extrusion.material
	matBack = extrusion.back_material
	
	#Prepare path
	if closed
		#Closed path, make the point along path closest to the face the start point (not limited to corners)
		#Closest to face means closest to first vertex returned #NOTE: use center of gravity?
		
		#Declare vars for closest point, distance and point index where to add the new point to path array
		#Start with first point
		indexOfClosest = 0
		pClosest = path[0]
		dClosest = extrusion.vertices[0].position.distance path[0]
		
		#Loop all segments
		(0..(path.length-2)).each do |i|
			#Get point on segment that is closest to face's first vertex
			p1 = path[i]
			p2 = path[i+1]
			line = [p1,(p2.-p1)]
			p = extrusion.vertices[0].position.project_to_line line
			#Check if point is on actual segment
			if(p1.distance(p) > p1.distance(p2))
				#Point is beyond p2, use p2 instead
				p = p2
			elsif(p2.distance(p) > p2.distance(p1))
				#Point is beyond p1, use p1 instead
				p = p1
			end#if elsif
			
			#Check if this point is closer than the closest one seen so far
			d = extrusion.vertices[0].position.distance p
			next if d > dClosest
			
			indexOfClosest = i
			pClosest = p
			dClosest = d
		end#each
		
		#Add closest point to path array if not already there (= if not a corner)
		unless(path.include? pClosest) 
			indexOfClosest+= 1
			path.insert(indexOfClosest, pClosest)
		end#unless
		
		#Temporary remove last point in path since it's the same as the first
		path.pop
		#Rotate array so start point is first
		(1..indexOfClosest).each do
			path << path[0]
			path.shift
		end#each
		#Copy first point to end of path so it again start and ends at the same place
		path << path[0]
	else
		#Open path, make path end closest to extrusion the start point
		#Closest to face means closest to first vertex returned #NOTE: use center of gravity?
		if ((extrusion.vertices[0].position.distance path[-1]) < (extrusion.vertices[0].position.distance path[0]))
			path.reverse!
		end#if
	end#if
	
	#Points around face to extrude
	startProfilePoints = []
	extrusion.outer_loop.vertices.each do |i|
		startProfilePoints << i.position
	end#each
	
	#Smooth points, indexes of points between curve segments
	smoothPoints = []
	(0..(extrusion.outer_loop.vertices.length-1)).each do |i|
		vertex = extrusion.outer_loop.vertices[i]
		smoothPoints << i if vertex.edges[0].curve != nil and vertex.edges[1].curve != nil
	end#each
	
	#Points of previous profile, used when drawing edges between profiles
	previousProfilePoints = startProfilePoints
	
	#Edges of previous profile, will be removed if co-planar
	#Starts as empty array since edges of first profile shouldn't be soften
	previousProfileEdges = []
	
	#Find first vector of path, used for rotation
	if closed
		#Closed path, use flatten average of vectors along segment before and after point
		vector1 = flattenVector((path[1].- path[0]), [Geom::Point3d.new, vectorUpright]).normalize
		vector2 = flattenVector((path[0].- path[-2]), [Geom::Point3d.new, vectorUpright]).normalize
		vectorRotationStart = Geom::linear_combination 1, vector1, 1, vector2
	else
		#Not closed path, use flatten vector along first segment
		vectorRotationStart = flattenVector((path[1].- path[0]), [Geom::Point3d.new, vectorUpright]).normalize
	end#if elsif
	
	#Find angle in start point if closed oath, used for scaling
	if closed
		angeInStartPoint = flattenVector((path[0].- path[-2]), [Geom::Point3d.new, vectorUpright]).angle_between flattenVector((path[1].- path[0]), [Geom::Point3d.new, vectorUpright])
		scaleStart = 1/(Math.cos(angeInStartPoint/2))
	end
	
	#If closed path, remove extrusion face
	extrusion.erase! if closed
	
	#Unless closed path, remove extrusion face if all its edges bind other faces (consistency with push-pull & followme)
	unless closed
		removeStartFace = true
		extrusion.outer_loop.edges.each do |j|
			removeStartFace = false unless j.faces.length == 2
		end#each
		extrusion.erase! if removeStartFace
	end#unless
	
	#Loop path each point but first (it's profile is already drawn since it's the face being extruded)
	#Draw a profiled on each point aligned to both the path and the upright vector
	#Connect this profile with previous profile
	(1..(path.length-1)).each do |i|
	
		#Translate transformation
		vectorMove = path[i].- path[0]
		translation = Geom::Transformation.translation vectorMove
		
		#Rotate transformation
		if i == (path.length-1)
			#Last point
			if closed
				#Closed path
				vectorRotation = vectorRotationStart
			else
				#Not closed path
				vectorRotation = flattenVector((path[i].- path[i-1]), [Geom::Point3d.new, vectorUpright]).normalize
			end#if
		else
			#Not last nor first point
			vector1 = flattenVector((path[i].- path[i-1]), [Geom::Point3d.new, vectorUpright]).normalize
			vector2 = flattenVector((path[i+1].- path[i]), [Geom::Point3d.new, vectorUpright]).normalize
			vectorRotation = Geom::linear_combination 1, vector1, 1, vector2
		end#if
		#Find angle to rotate(positive engle returned)
		rotationAngle = vectorRotation.angle_between vectorRotationStart
		#Check if angle should be negative
		negAnglePointForward = Geom::Point3d.new.offset vectorRotationStart
		negAnglePointPos = negAnglePointForward.transform(Geom::Transformation.rotation(Geom::Point3d.new, vectorUpright, Math::PI/2))
		negAnglePointNeg = negAnglePointForward.transform(Geom::Transformation.rotation(Geom::Point3d.new, vectorUpright, -Math::PI/2))
		negAnglePointAlongPath = Geom::Point3d.new.offset vectorRotation
		rotationAngle*= -1 if (negAnglePointAlongPath.distance negAnglePointNeg) < (negAnglePointAlongPath.distance negAnglePointPos)
		#Create transformation
		rotation = Geom::Transformation.rotation path[i], vectorUpright, rotationAngle
		
		#Scale transformation
		#To keep the sides of the extrusion parallel (consistency with followme)
		#Done as several subtransformation that are later merged to the actual scaling transformation
		#angle path turns in this point, projected to plane perpendicular to vectorUprigh, doesn't matter if positive or negative
		if i == (path.length-1)
			#Last point
			if closed
				#Closed path
				angeInPoint = angeInStartPoint
			else
				#Not closed path
				angeInPoint = 0
			end#if
		else
			#Not last nor first point
			angeInPoint = flattenVector((path[i].- path[i-1]), [Geom::Point3d.new, vectorUpright]).angle_between flattenVector((path[i+1].- path[i]), [Geom::Point3d.new, vectorUpright])
		end
		scale = 1/(Math.cos(angeInPoint/2))
		#Divide scale with scale of first profile. If first profile is in corner its width wont be the with on extrusion between corners
		scale/= scaleStart if closed
		#create scale transformation from origin in x axis
		scaling = Geom::Transformation.scaling Geom::Point3d.new, scale, 1, 1
		#Rotate scale transformation
		rScaling = Geom::Transformation.axes path[i], (vectorRotation.cross vectorUpright), vectorRotation, vectorUpright
		scaling = rScaling.*scaling.*(rScaling.inverse)
		
		#Perform translation
		profilePoints = []
		startProfilePoints.each do |j|
			profilePoints << j.transform((scaling.*rotation).*(translation))
		end
		
		#Draw profile
		#NOTE: why draw faces?
		face = ents.add_face profilePoints
		pofilesEdges = face.edges
		face.erase!
		
		#Connect corresponding points to previous profile
		smoothExtrudeLines = []
		(0..(profilePoints.length-1)).each do |j|
			edgeStraight = ents.add_line profilePoints[j], previousProfilePoints[j]
			smoothExtrudeLines << edgeStraight if smoothPoints.include? j
		end#each
		
		#Draw diagonals and form faces towards previous profile
		diagonals = []
		(0..(profilePoints.length-1)).each do |j|
			#NOTE: better name for nextIndex
			nextIndex = j+1
			nextIndex = 0 if nextIndex == profilePoints.length
			edge = ents.add_line profilePoints[j], previousProfilePoints[nextIndex]
			diagonals << edge
			#Draw each face at a time instead of using edge.find_faces. The face connected to previous segment should be drawn first to prevent back face problems.
			ents.add_face profilePoints[j], previousProfilePoints[j], previousProfilePoints[nextIndex]
			ents.add_face profilePoints[j], profilePoints[nextIndex], previousProfilePoints[nextIndex]
			#Paint new faces
			edge.faces.each do |k|
				k.material = mat
				k.back_material = matBack
			end#each
		end#each
		
		#Soften and remove co-planar edges in diagonals, previous profile, and edges extruded from curve
		#NOTE: Add a separate array of what points in path should have its profile soften, soften if one of the edges connected to the point is a part of a curve (like native follow me)? 
		#NOTE: extrude() could be called with second parameter for softening profiles. All hard, all soft or array of soft corners. userInput() could set this depending on if vertices holds edges that are part of curves. modifier keys for all/none soften.
		(diagonals + previousProfileEdges+ smoothExtrudeLines).each do |j|
			removeOrSoftenSmooth(j)
		end#each
		
		#Save data for next iteration
		previousProfilePoints = profilePoints
		previousProfileEdges = pofilesEdges
	end#each
	
	#If closed path, soften first/last profile's edges
	if closed
		previousProfileEdges.each do |i|
			removeOrSoftenSmooth(i)
		end#each
	end#if

	#Add end face unless all its edges bind an outer face or path is closed(consistency with push-pull & followme)
	unless closed
		drawEndFace = false
		previousProfileEdges.each do |i|
			drawEndFace = true if i.faces.length != 2
		end#each
		if drawEndFace
			endFace = ents.add_face previousProfileEdges
			endFace.material = mat
			endFace.back_material = matBack
		end#if
	end#unless
	
	#Return end face. Can be used by other plugins calling the extruder
	return endFace if endFace
	
end#def

end#module